<?php

namespace loyalsoft;

/**
 * 错误级别
 */
class enum_LogLevel extends Enum {

    const All = 0;
    const Info = 1;
    const Warn = 2;
    const Err = 4;

}

/**
 * 日志工具类_高性能(High Performance)
 * @author gwang (mail@wanggangzero.cn)
 * @version <br/>
 *          3.0.2 2018年3月9日 完善稳定性, GetDir()方法添加了自动检查目录是否存在的逻辑. -gwang <br/>
 *          3.0.1 2018年1月9日 整理, online情况下paylog->mysql, errlog->redis, log和info信息仍然在文件(online默认关闭log和info的落地操作). -gwang <br/>
 *          2.0.1 2017年8月8日 将paylog和giftlog和output也保留了.以前向兼容就代码. -gwang <br/>
 *          2.0   2017年8月8日 整理, 将redis部分暂时去掉, 保留缓存模式, 在处理的最后集中flush到文件中. -gwang <br/>
 *          1.0   2016年7月1日 创建, 思路: 建立高性能的日志模块,消除写文件.暂时不包含支付日志. -gwang <br/>
 */
class CLog {

    /**
     * 日志缓存
     * @var array
     */
    private static $Warn = array();
    private static $Info = array();
    private static $Err = array();
    private static $logLevel = enum_LogLevel::Err;

    /**
     * @var int 日志写入redis时保存的记录条数
     */
    const redis_log_trim_max = 5000;

    /**
     * @var int 日志写入redis时每次trim的记录条数
     */
    const redis_log_trim_once = 100;

    /**
     * @return string 日志目录
     */
    private static function GetDir($subfolder = null) {
        # 线上版的时候会是linux系统, 单独指定目录, 测试的时候就放到代码旁边得了
        $dir = (GAME_ONLINE ? "/data/" : ROOTDIR . "/../" ) . "logs/" . PROJECTNAME . "/";
        if (!empty($subfolder)) {                                               # 如果有子目录, 附加
            $dir .= $subfolder . "/";
        }
        if (is_dir($dir) || mkdir($dir, 0777, true)) {                          # 如果目录不存在, 创建
            return $dir;
        }
        exit("can not access log directory!");
    }

    /**
     * 设置日志级别
     * @param int $level 参考常量enum_LogLevel :: 0 => info, 1 => log, 2 => err
     */
    public static function SetLogLevel($level) {
        if ($level >= enum_LogLevel::All && $level <= enum_LogLevel::Err) {
            self::$logLevel = $level;
        } else {
            // 无效值,忽视
        }
    }

    /**
     * 断言日志,注意此断言不是打断程序执行的那种.
     * 如果断言失败,则会在日志里面留下一条Err记录,记录内容为errmsg.
     * 如果判断成功,且设置了okmsg, 则会在日志里面留下一条info记录记录内容为OKmsg.
     * @param bool $condition 条件/条件表达式
     * @param string $errmsg 条件不成立时输出到日志中的信息, 日志级别err
     * @param string $okmsg  【可选】默认(null), 当条件成立时不记录任何消息, 除非特别指定了消息内容.日志级别info
     */
    public static function Assert($tag, $condition, $errmsg, $okmsg = null) {
        if (!$condition) {
            self::err($errmsg, $tag);
        } else if ($okmsg) {
            self::info($okmsg, $tag);
        }
    }

    /**
     * 记录日志
     * @param string $msg
     * @param string $tag 标签
     */
    public static function warn($msg, $tag = 'log') {
        if (self::$logLevel <= enum_LogLevel::Warn) {
            self::$Warn[] = self::makeLogMsg($msg, $tag);
        }
    }

    /**
     * @param string $msg
     * @param string $tag 标签
     */
    public static function info($msg, $tag = 'info') {
        if (self::$logLevel <= enum_LogLevel::Info) {
            self::$Info[] = self::makeLogMsg($msg, $tag);
        }
    }

    /**
     * 记录异常日志
     * @param string $msg
     * @param string $tag 标签
     */
    public static function err($msg, $tag = "err") {
        if (self::$logLevel <= enum_LogLevel::Err) {
            self::$Err[] = self::makeLogMsg($msg, $tag);
        }
    }

    /**
     * 将缓存中的数据写入存储设备并清除缓存
     */
    public static function flush() {
        self::flush2Redis(enum_LogLevel::Info);
        self::flush2Redis(enum_LogLevel::Warn);
        self::flush2Redis(enum_LogLevel::Err);
    }

// <editor-fold defaultstate="collapsed" desc="辅助方法">

    /**
     * 将日志写入文件
     * @global type $zoneid
     * @param enum_LogLevel $type
     */
    private static function flush2File($type) {
        $typename = strval(new enum_LogLevel($type));
        $fileName = self::GetDir() . $typename . "-" . date('Ymd') . ".log";    # 日志文件名(按天分割)
        $fd = fopen($fileName, "a");
        if (false === $fd) {                                                    # 打开文件失败
            throw new \Exception("打开 $fileName 失败");
        }
        $arr = self::$$typename;                                                # 取对应类型的数组
        if (count($arr) > 0) {
            foreach ($arr as $msg) {
                $n = fputs($fd, $msg . PHP_EOL);
                if (false === $n) {                                             # 写入时失败
                    throw new \Exception("写入 $fileName 时失败");               # 这种的就得靠查找系统(nginx/php-fpm)日志了
                }
            }
            self::$$typename = array();                                         # 清空数据
        }
        fclose($fd);                                                            # 关闭文件
    }

    /**
     * 将缓存中的数据写入存储设备(redis)并清除缓存
     * @type int/enum_LogLevel 日志类型
     */
    private static function flush2Redis($type) {
        $typename = strval(new enum_LogLevel($type));
        $key = "log-" . $typename;
        $arr = self::$$typename;                                                # 取对应类型的数组
        if (count($arr) > 0) {
            if (null != req()) {
                $arr[] = "" . req();
            }
            gMem()->lpush($key, $arr);
            if (gMem()->llen($key) > self::redis_log_trim_max + self::redis_log_trim_once) {# 达到清理条件
                gMem()->ltrim($key, 0, -self::redis_log_trim_once);             # 缩减记录
            }
            self::$$typename = array();                                         # 清空数据
        }
    }

    /**
     * 将缓存中的数据写入存储设备(mysql)并清除缓存
     * @type int/enum_LogLevel 日志类型
     */
    private static function flush2MySQL($type) {
        $typename = strval(new enum_LogLevel($type));
        $key = "log-" . $typename;
        $arr = self::$$typename;                                                # 取对应类型的数组

        if (count($arr) > 0) {
            $sql = "create table  if not exists `$key` (`row` INT(11) NOT NULL AUTO_INCREMENT COMMENT '自增行号', `msg` TEXT NULL COMMENT '错误消息内容',	PRIMARY KEY (`row`)) COLLATE='utf8_general_ci' ENGINE=InnoDB;";
            daoInst()->exec($sql);                                              # 创建表
            $data = array();
            array_map(function ($v) use (&$data) {
                $s = daoInst()->quote($v);
                $data[] = "(" . $s . ")";
            }, $arr);
            $sql_insert = "insert into `$key` (`msg`) values " . implode(',', $data) . " ;";
            daoInst()->exec($sql_insert);                                       # 执行插入操作
            $n = daoInst()->select()->from("`$key`")->count();                  # 查一下当前表大小
            $max = 150000;                                                      # 最多保留15万条记录
            $del_once = 10000;                                                  # 超过15万,清理一次,一次删除1万条
            if ($n > $max) {
                $sql = "delete from `$key` order by `row` limit $del_once ;";   # 一次清理一批
                daoInst()->exec($sql);
            }
            self::$$typename = array();                                         # 清空数据
        }
    }

    /**
     * 生成日志,格式: [x区][时间][文件][tag]: msg
     * @global \loyalsoft\type $zoneid
     * @param type $msg
     * @param type $tag
     * @return string
     */
    private static function makeLogMsg($msg, $tag = null) {
        global $zoneid;
        $array = debug_backtrace();
        while (isset($array[0]) && isset($array[0]['file']) && $array[0]['file'] == __FILE__) {
            array_shift($array);
        }
        $row = $array[0];
        $server_addr = isset($_SERVER['SERVER_ADDR']) ? $_SERVER['SERVER_ADDR'] : "-";
        $client_addr = isset($_SERVER['REMOTE_ADDR']) ? $_SERVER['REMOTE_ADDR'] : "-";
        $uid = empty(req()->uid) ? "-" : req()->uid;
        $str = "[" . $server_addr . "][$zoneid 区][" . $uid . "]"                          # 请求主机ip地址、分区, uid
                . "[" . date('Y-m-d H:i:s') . "][" . $client_addr . "]["                   # 时间, 客户端来源ip地址
                . (isset($row['file']) ? CommUtil::str2UTF8(basename($row['file'])) : '-') # 所在文件名
                . ":" . (isset($row['line']) ? $row['line'] : "???")#                      # 所在行
                . "]" . (isset($tag) ? "[" . CommUtil::str2UTF8($tag) . "]" : "")          # 标签
                . "=> " . var_export($msg, true)#                                          # 内容
                . PHP_EOL;
        return $str;
    }

// </editor-fold>
//
// <editor-fold defaultstate="collapsed" desc="优化后的文件日志(引入cache)">

    /**
     * 写入文件日志
     * @global \loyalsoft\type $zoneid
     * @param type $type
     * @param type $msg
     * @return type
     */
    private static function fileLog($type, $msg) {
        $fileName = self::GetDir($type) . date('Ymd') . ".log";                 # 日志文件名(按天分割)
        $fd = fopen($fileName, "a");
        if (is_array($msg) || is_object($msg)) {
            $msg = JsonUtil::encode($msg);
        }
        global $zoneid;
        fputs($fd, "[" . self::dtCurrent() . "] $zoneid 区, " . $msg . "\r\n");
        fclose($fd);
    }

    private static $cache = array();

    private static function output2($type, $msg) {
        if (!isset(self::$cache[$type])) {
            self::$cache[$type] = array();
        }
        if (is_array($msg) || is_object($msg)) {
            $msg = JsonUtil::encode($msg);
        }
        global $zoneid;
        self::$cache[$type][] = "[" . self::dtCurrent() . "] $zoneid 区, " . $msg . "\r\n";
    }

    private static function flushFileCache() {
        if (is_array(self::$cache)) {
            foreach (self::$cache as $t => $varr) {
                $fileName = self::GetDir($t) . date('Ymd') . ".log";            # 日志文件名(按天分割)
                $fd = fopen($fileName, "a");
                if (is_array($varr)) {
                    foreach ($varr as $msg) {
                        fputs($fd, $msg);
                    }
                }
                fclose($fd);
            }
            self::$cache = array();
        }
    }

// </editor-fold>
//
// 避免异常导致缓存丢失, 付费模块的日志直接写入文件
// <editor-fold defaultstate="collapsed" desc=" 付费相关日志, 无缓存 ">

    /**
     * 输出日志 到pay.log
     * @param mixed $msg 参数是字符串则直接输出,否则调用json_encode
     */
    public static function pay($msg) {
        $str = self::makeLogMsg($msg);                                          # 统一日志格式字符串
        self::Paylog2Mysql($str);
//        if (GAME_ONLINE) {
//            self::Paylog2Mysql($str);
//        } else {
//            $fileName = self::GetDir() . "pay" . date('Ym') . ".log";
//            self::Log2File($fileName, $str);
//        }
    }

    /**
     * 向文件写入日志
     * @global int $zoneid
     * @param string $fileName
     * @param string $msg
     */
    private static function Log2File($fileName, $msg) {
        $fd = fopen($fileName, "a");
        if ($fd === FALSE) {                                                    # 打开文件失败
            throw new \Exception('open file failed! ' . $fileName);
        }
        fputs($fd, $msg . PHP_EOL);
        fclose($fd);                                                            # 关闭文件
    }

    /**
     * 将pay日志写入mysql
     * @param type $msg
     */
    private static function Paylog2Mysql($msg) {
        $t = "tab_paylog_" . date('Ym');
        $sql = "create table if not exists $t like tpl_paylog_tab_ym;"          # 组装SQL
                . " insert into $t (msg) values ("
                . daoInst()->quote($msg) . ");";                                # 消息体
        $n = daoInst()->exec($sql);
        if ($n < 0) {
            self::err("写入数据库失败." . $sql);
        }
    }

// </editor-fold>
//
}

if (!GAME_ONLINE) {                                                             # 除了现网
    CLog::SetLogLevel(enum_LogLevel::All);                                      # 开启所有级别的日志输出
}